local MAJOR_VERSION = "Threat-1.0"
local MINOR_VERSION = tonumber(("$Revision: 41446 $"):match("%d+"))

if MINOR_VERSION > _G.ThreatLib_MINOR_VERSION then
	_G.ThreatLib_MINOR_VERSION = MINOR_VERSION
end

ThreatLib_funcs[#ThreatLib_funcs+1] = function()

---------------------------------------------------------------------------
-- Global stuff
---------------------------------------------------------------------------

local _G = _G
local tonumber = _G.tonumber
local type = _G.type
local select = _G.select
local next = _G.next
local pcall = _G.pcall
local pairs = _G.pairs
local tinsert = _G.tinsert
local UnitExists = _G.UnitExists
local UnitName = _G.UnitName
local UnitClass = _G.UnitClass
local UnitCanAttack = _G.UnitCanAttack
local GetNumRaidMembers = _G.GetNumRaidMembers
local GetNumPartyMembers = _G.GetNumPartyMembers

---------------------------------------------------------------------------
-- ThreatLib and its internals
---------------------------------------------------------------------------

local ThreatLib = _G.ThreatLib

local partyMemberUsesKLHTM = ThreatLib.partyMemberUsesKLHTM
local partyMemberRevisions = ThreatLib.partyMemberRevisions
local IsGroupOfficer = ThreatLib.IsGroupOfficer
local partyUnits = ThreatLib.partyUnits
local TailoredBinaryCheckSum = ThreatLib.TailoredBinaryCheckSum

local playerName = UnitName("player")

---------------------------------------------------------------------------
-- KTM message processing
---------------------------------------------------------------------------

local OnPureKLHTMMessage = {}
local OnAnyKLHTMMessage = {}

local function handleMessage(self, handlers, sender, command, data)
	local func = handlers[command]
	if func then
		func(self, sender, data)
	end
end

function ThreatLib:CHAT_MSG_ADDON(prefix, message, distribution, sender)
	if prefix == "KLHTM" and (distribution == "PARTY" or distribution == "RAID") then
		self:Debug("Received from KLHTM: %q", message)
		local command, data = message:match("^(%a+) ?(.*)$")
		if command then			
			if sender ~= playerName then
				if not partyMemberUsesKLHTM[sender] then
					partyMemberUsesKLHTM[sender] = true
					self:TriggerEvent('Threat_KLHTM_NewPeer', sender)			
				end

				if partyMemberUsesKLHTM[sender] and not partyMemberRevisions[sender] then
					handleMessage(self, OnPureKLHTMMessage, sender, command, data)
				end
			end
			handleMessage(self, OnAnyKLHTMMessage, sender, command, data)
		end
	end
end

---------------------------------------------------------------------------
-- Threat update and misdirection
---------------------------------------------------------------------------

function OnPureKLHTMMessage:t(sender, data)
	local _ = partyUnits[nil] -- refresh if need be
	local threat = tonumber(data)
	if threat then
		self:ThreatUpdatedForUnit( sender, TailoredBinaryCheckSum[self.KLHTMtarget or "_GLOBAL_"], threat)
	end
end

function OnPureKLHTMMessage:misdirection(sender, data)
	local targetPlayer, amount = data:match("^(%S+) (%d+)$")
	if targetPlayer == playerName and select(2, UnitClass(partyUnits[sender])) == "HUNTER" then
		local targetEnemy = self.KLHTMtarget or "_GLOBAL_"
		amount = tonumber(amount)
		self:Debug("Misdirected threat from KLHTM: %q (%q)", amount, targetEnemy)
		self:GetModule("ClassCore").activeModule:_addTargetThreat(amount, targetEnemy)
	end
end

---------------------------------------------------------------------------
-- Master target
---------------------------------------------------------------------------

local function UpdateKLHTMTarget(target)
	if (target or "") ~= (ThreatLib.KLHTMtarget or "") then
		ThreatLib.KLHTMtarget = target
		ThreatLib:TriggerEvent('Threat_KLHTM_TargetChanged', target)
	end
end

function OnAnyKLHTMMessage:target(sender, data)
	if data and type(data) == "string" then
		local senderUnit = partyUnits[sender]
		if IsGroupOfficer(senderUnit) then
			local senderTargetUnit = senderUnit..'target'
			if UnitExists(senderTargetUnit) then
				UpdateKLHTMTarget(UnitName(senderTargetUnit))
			else
				UpdateKLHTMTarget(self:TranslateForeignMobName(data))
			end
			self:Debug("KTM Target set to %s (%s) by %s", self.KLHTMtarget, data, sender)
		else
			self:Debug("Dropped target from %q", sender)
		end
	end
end

function OnAnyKLHTMMessage:cleartarget(sender, data)
	local senderUnit = partyUnits[sender]
	if IsGroupOfficer(senderUnit) then
		UpdateKLHTMTarget(nil)
		self:Debug("KTM Target cleared by %s", sender)
	else
		self:Debug("Dropped cleartarget from %q", sender)
	end
end

---------------------------------------------------------------------------
-- Version responses
---------------------------------------------------------------------------

function OnPureKLHTMMessage:versionresponse(sender, version)
	local senderUnit = partyUnits[sender]
	if senderUnit and version:match("^%d+\.%d+$") and not partyMemberRevisions[sender]  then
		if version ~= (partyMemberUsesKLHTM[sender] or false) then
			partyMemberUsesKLHTM[sender] = version
			self:TriggerEvent('Threat_KLHTM_VersionResponse', sender, version)
		end
	end
end

---------------------------------------------------------------------------
-- Empty KLHTM output methods
---------------------------------------------------------------------------

do
	local function noop() end

	ThreatLib.IsKLHTMSupportLoaded = noop
	ThreatLib.IsKLHTMBroadcastEnabled = noop
	ThreatLib.IsKLHTMBroadcastActive = noop

	ThreatLib.PublishThreatToKLHTM = noop
	ThreatLib.MisdirectToKLHTM = noop
	
	ThreatLib.SetKLHTMMasterTarget = noop
	ThreatLib.ClearKLHTMMasterTarget = noop
	ThreatLib.QueryKLHTMVersions = noop

	ThreatLib.SendKLHTMMessage = noop
end

---------------------------------------------------------------------------
-- Real KLHTM output methods
---------------------------------------------------------------------------

local function LoadKLHTMBroadcastSupport()

	-- Won't load if the real KLHTM has been loaded
	if _G.klhtm then
		return
	end
	
	-- Locals
	local fakeKLHTMVersion = "19.18"
	local lastPublishedThreat
	local enabled

---------------------------------------------------------------------------
-- Status accessors
---------------------------------------------------------------------------
	
	function ThreatLib:IsKLHTMSupportLoaded()
		return true
	end
	
	function ThreatLib:EnableKLHTMBroadcast(value)
		if enabled and not value then
			if lastPublishedThreat and lastPublishedThreat > 0 then
				self:PublishThreatToKLHTM(nil)
			end
			enabled = nil
			self:TriggerEvent('Threat_KLHTM_BroadcastDisabled')
			
		elseif not enabled and value then
			enabled = true
			lastPublishedThreat = nil
			self:TriggerEvent('Threat_KLHTM_BroadcastEnabled')
		end
	end
		
	function ThreatLib:IsKLHTMBroadcastEnabled()
		return enabled
	end
	
	function ThreatLib:IsKLHTMBroadcastActive()
		return enabled and next(partyMemberUsesKLHTM) and self:IsActive()
	end
	
---------------------------------------------------------------------------
-- Publishing methods
---------------------------------------------------------------------------
	
	function ThreatLib:PublishThreatToKLHTM(playerHash)
		if not enabled then return end
		local threat 
		if playerHash and next(playerHash) then
			if self.KLHTMtarget then
				threat = playerHash[TailoredBinaryCheckSum[self.KLHTMtarget]]
			else
				threat = 0
				for k,v in pairs(playerHash) do
					threat = threat + v
				end
			end
		else
			threat = 0
		end
		if threat and threat ~= (lastPublishedThreat or 0) then
			if self:SendKLHTMMessage('t', threat) then
				lastPublishedThreat = threat
			end
		end
	end
	
	function ThreatLib:MisdirectToKLHTM(targetPlayer, threat)
		if not enabled then return end		
		self:SendKLHTMMessage('misdirection', ("%s %d"):format(targetPlayer, threat))
	end
	
	function ThreatLib:SetKLHTMMasterTarget(targetUnit)
		if not enabled or not IsGroupOfficer("player") then
			self:Debug("Group priviledge required")
			return
		end
		if not targetUnit then
			targetUnit = "target"
		end
		if UnitExists(targetUnit) and UnitCanAttack("player", targetUnit) then
			local target = UnitName(targetUnit)
			if self:SendKLHTMMessage("target", target) then
				UpdateKLHTMTarget(target)
				return true
			end
		end
	end
	
	function ThreatLib:ClearKLHTMMasterTarget()
		if not enabled or not IsGroupOfficer("player") then
			self:Debug("Group priviledge required")
			return
		end
		if self:SendKLHTMMessage("cleartarget") then
			UpdateKLHTMTarget(nil)
			return true
		end
	end	
	
	function ThreatLib:QueryKLHTMVersions()
		if not IsGroupOfficer("player") then
			self:Debug("Group priviledge required")
			return
		end
		return self:SendKLHTMMessage("versionquery", nil, true)
	end	
	
	function OnPureKLHTMMessage:versionquery(sender, data)
		if enabled and IsGroupOfficer(partyUnits[sender]) then
			self:SendKLHTMMessage("versionresponse", fakeKLHTMVersion)
		end
	end
	
---------------------------------------------------------------------------
-- Actual sending method
---------------------------------------------------------------------------

	function ThreatLib:SendKLHTMMessage(command, data, force)
		local _ = partyUnits[nil] -- refresh if need be
		if not enabled or (not force and not next(partyMemberUsesKLHTM)) then -- no point
			return
		end
		local distribution
		if GetNumRaidMembers() > 0 then
			distribution = "RAID"
		elseif GetNumPartyMembers() > 0 then
			distribution = "PARTY"
		else
			return
		end
		
		local message
		if data then
			message = command .. " " .. data
		else
			message = command
		end
		
		ChatThrottleLib:SendAddonMessage("NORMAL", "KLHTM", message, distribution)
		self:Debug("Sent to KLHTM: %q", message)
		return true
	end
		
---------------------------------------------------------------------------
-- Activation
---------------------------------------------------------------------------	

	ThreatLib:TriggerEvent('Threat_KLHTM_BroadcastLoaded')
	ThreatLib:EnableKLHTMBroadcast(true)
	
end

---------------------------------------------------------------------------
-- Enabling stub
---------------------------------------------------------------------------

-- This method will be overriden by LoadKLHTMBroadcastSupport()
function ThreatLib:EnableKLHTMBroadcast(value)
	if value then
		local res, reason = pcall(LoadKLHTMBroadcastSupport)
		if not res then
			geterrorhandler()(reason)
		end
	end
end

---------------------------------------------------------------------------

tinsert(ThreatLib.UpvalueFixers, function(lib, oldLib)
	ThreatLib = lib
	if oldLib and oldLib:IsKLHTMBroadcastEnabled() then
		ThreatLib:EnableKLHTMBroadcast(true)
	end
end)

end

